#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>

#include "../mimer/defs.h"
#include "../mimer/data.h"
#include "../mimer/eurist.h"
#include "../mimhelp/mimhelp.h"
#include "mimpatch.h"

//---------------------------------------------------------------------------

#define VERSION "MimPATCH for " MIMER_VERSION

#define ACTION_INFO     	1
#define ACTION_MODIFY   	2
#define ACTION_MOVE     	4
#define ACTION_PATCH_HELP	8

void   __crt0_load_environment_file(char *_app_name)
{;}

char mh_patchblock[2*MH_PATH_MAX+sizeof(MH_SIGNATURE)]=MH_SIGNATURE;

void usage(void)
{
fputs(VERSION "General options:\n"
             "\t-f    <file to be patched>\n"
             "\t-i\t\t\t\t\tprint info\n"
             "\t-F    <file to get patch from>\n"
             "\t-r    <reliable charset list>\n"
             "\t-u    <unreliable charset list>\n"
             "\t-g|-G <file to gather statistics from>\twrite|append\n"
             "\t-c|-C <charset file>\t\t\tadd|remove charset\n"
             "\t-H    <path to MimHelp.EXE file>\n"
             "\t-x    <extension for ouput files>\n"

             "Mimer options:\n"
             "\t-d|-D\t\t\t\t\tdelete input|don't delete\n"
             "\t-h    <custom header>\n"
             "\t-o    <destination directory>",stderr);
             
exit(-1);
}

int memem(char *buf, char *mask, int n)
{
int x=strlen(mask)+1;
int i;

n-=x;

for (i=0;i<n;i++)
    if (memcmp(buf+i,mask,x)==0)
       return(i);
return(-1);
}

off_t find_signature(int fd, char *sig)
{
const BUFSIZE=256;

char buf[2*BUFSIZE];
off_t offset=0;
int   p;

memset(buf,0,sizeof(buf));
lseek(fd,0,SEEK_SET);

read(fd,buf,BUFSIZE*2);

while ((p=memem(buf,sig,2*BUFSIZE))<0)
      {
      memcpy(buf,buf+BUFSIZE,BUFSIZE);
      if (read(fd,buf+BUFSIZE,BUFSIZE)<=0)
         return(-1);
      offset+=BUFSIZE;
      }
return(offset+p);
}

off_t read_patch(char *path, patch * dst, int perr)
{
off_t pos;
int fdi=open(path,O_RDONLY|O_BINARY);

if (fdi<0)
   {
   if (perr)
      {
      perror("Unable to open input file");
      return(-1);
      }
   return(0);
   }

if ((pos=find_signature(fdi,MIMER_SIGNATURE))<0)
   {
   fprintf(stderr,"File %s doesn't seem to contain patch information",path);
   return(-1);
   }

lseek(fdi,pos,SEEK_SET);
if (read(fdi,dst,sizeof(patch))!=sizeof(patch))
   {
   perror("Can't read patch information");
   return(-1);
   }

close(fdi);
return(pos);
}

int main(int argc, char *argv[])
{
unsigned int action=0;
off_t        pos;
int          fd;
char         *path=NULL;
patch        data;     //={MIMER_SIGNATURE};
char         c;
int          i;
char	     *hpath=NULL;

memset(&data,0,sizeof(data));
strcpy(data.signature,MIMER_SIGNATURE);
data.charset_reliability=RELIABLE;
data.factor=1.;

action|=ACTION_MODIFY;

while ((c=getopt(argc, argv, "f:iF:r:u:g:G:c:C:H:dDh:o:x:")) != -1)
      switch(c)
        {
        case 'g':
           memset(data.frequencies,0,sizeof(data.frequencies));
        case 'G':
           {
           eurist e=&data;

           if (e.gather(optarg)<0)
              exit(-1);
           action|=ACTION_MODIFY;
           }
           break;

	case 'H':
           hpath=optarg;
           action|=ACTION_PATCH_HELP;
	   break;
           
        case 'd':
           data.dflag=1;
           action|=ACTION_MODIFY;
           break;

        case 'D':
           data.dflag=0;
           action|=ACTION_MODIFY;
           break;

        case 'x':
           if (*optarg=='.')
           	strncpy(data.dst_xtension,optarg,PATCH_DST_XTN_SIZE-1);
           else
           	{
           	strncpy(data.dst_xtension+1,optarg,PATCH_DST_XTN_SIZE-2);
                data.dst_xtension[0]='.';
                }
           data.dst_xtension[PATCH_DST_XTN_SIZE-1]=0;
           action|=ACTION_MODIFY;
           break;

        case 'h':
           strncpy(data.xheader,optarg,PATCH_XHEADER_SIZE-2);
           data.xheader[PATCH_XHEADER_SIZE-2]=0;
           action|=ACTION_MODIFY;
           break;

        case 'o':
           strncpy(data.dst_path,optarg,PATH_MAX-1);
           data.dst_path[PATH_MAX-1]=0;
           action|=ACTION_MODIFY;
           break;

        case 'r':
           strncpy(data.charset_regex,optarg,PATCH_RE_CHARSET_LIST_SIZE-1);
           data.charset_regex[PATCH_RE_CHARSET_LIST_SIZE-1]=0;
	   if (testre(data.charset_regex))
           	exit(-1);
           data.charset_reliability=RELIABLE;
           action|=ACTION_MODIFY;
           break;

        case 'u':
           strncpy(data.charset_regex,optarg,PATCH_RE_CHARSET_LIST_SIZE-1);
           data.charset_regex[PATCH_RE_CHARSET_LIST_SIZE-1]=0;
	   if (testre(data.charset_regex))
           	exit(-1);
           data.charset_reliability=UNRELIABLE;
           action|=ACTION_MODIFY;
           break;

        case 'i':
           action|=ACTION_INFO;
           break;

        case 'F':
           if (read_patch(optarg,&data,1)<0)
              exit(-1);
           action&=~ACTION_MODIFY;
           action|=ACTION_MOVE;
           break;

        case 'f':
           if (path!=NULL)
              {
              fputs("Several -f commands... What for?",stderr);
              exit(-1);
              }

           if ((pos=read_patch(optarg,&data,0))<0)
              exit(-1);

           path=optarg;
           action&=~ACTION_MODIFY;

           break;

        case 'C':
           for (i=0;i<data.charset_number;i++)
               if (strcmp(data.charset_name[i],optarg)==0)
                  break;

           if (i>=data.charset_number)
              {
              fprintf(stderr,"Can't find chrset table %s",optarg);
              exit(-1);
              }

           action|=ACTION_MODIFY;
           if (i==--data.charset_number)
              break;
           strcpy(data.charset_name[i],data.charset_name[data.charset_number]);
           memcpy(data.charset_table[i],data.charset_table[data.charset_number],256);
           break;

        case 'c':
           for (i=0;i<data.charset_number;i++)
               if (strcmp(data.charset_name[i],optarg)==0)
                  break;

           if (i>=PATCH_CHARSET_NUMBER)
              {
              fputs("Too many charset tables",stderr);
              exit(-1);
              }


           if ((fd=open(optarg,O_RDONLY|O_BINARY))<0)
              {
              perror("Can't open charset file");
              exit(-1);
              }

           if (read(fd,data.charset_table[i],256)<256)
              {
              perror("Can't read charset file");
              exit(-1);
              }
           close(fd);
           
           if (i>=data.charset_number)
              data.charset_number++;
           strncpy(data.charset_name[i],optarg,PATCH_CHARSET_NAME_SIZE-1);
           data.charset_name[i][PATCH_CHARSET_NAME_SIZE-1]=0;
           action|=ACTION_MODIFY;
           break;
              
        case '?':
           usage();
        }

if ((optind<argc) || (path==NULL))
   usage();

if (action&ACTION_MODIFY)
   data.time_stamp=time(NULL);

if (action&(ACTION_MODIFY|ACTION_MOVE))
   {
   if ((fd=open(path,O_WRONLY|O_CREAT|O_BINARY,S_IWUSR))<0)
       {
       perror("Can't open destination");
       exit(-1);
       }
   lseek(fd,pos,SEEK_SET);

   if (write(fd,&data,sizeof(patch))!=sizeof(patch))
       {
       perror("Can't write destination");
       exit(-1);
       }
   close(fd);
   }

if (action&ACTION_INFO)
   {
   printf("\n%s",VERSION);
   printf("\nFile \"%s\" contains following information:",path);
   printf("\n\tvalid signature \"%s\" is found at position %#X",data.signature, pos);
   printf("\n\tcustom-header line is ");
   if (*data.xheader)
      printf("\"%s\"",data.xheader);
   else
      printf("empty");
   printf("\n\tinput files are %sdeleted\n\t",(data.dflag ? "" : "not "));
   if (data.charset_reliability==UNRELIABLE)
      printf("un");
   printf("reliable charset list is \"%s\"",data.charset_regex);

   printf("\n\tavailable translation tables are <");
   for (i=0;i<data.charset_number;i++)
       printf("%s ",data.charset_name[i]);

   printf(">\n\toutput path is ");
   if (*data.dst_path)
      printf("\"%s\"",data.dst_path);
   else
      printf("not defined");
      
   printf("\n\tdefault extension for output files is \"%s\"",data.dst_xtension);

   printf("\n\tlast updated on %s", ctime(&data.time_stamp));
   }

if (action&ACTION_PATCH_HELP)
   {
   if (*data.dst_path==0)
       {
       fprintf(stderr,"Output path is unavailable - can't patch %s",hpath);
       exit(-1);
       }
   if (data.dst_path[strlen(data.dst_path)-1]!='\\' )
      strcat(data.dst_path,"\\");
   strcat(data.dst_path,"mimer.tsk");
   if (strlen(data.dst_path)>=MH_PATH_MAX)
       {
       fprintf(stderr,"Output path %s is too long for MimHelp.EXE",data.dst_path);
       exit(-1);
       }
   strcpy(mh_task_path,data.dst_path);

   if (strlen(path)>=MH_PATH_MAX)
       {
       fprintf(stderr,"Mimer path %s is too long for MimHelp.EXE",path);
       exit(-1);
       }
   strcpy(mh_exe_path,path);

   if ((fd=open(hpath,O_RDWR|O_BINARY))<0)
       {
       perror("Can't open MimHelp.EXE");
       exit(-1);
       }

   if ((pos=find_signature(fd,MH_SIGNATURE))<0)
      {
      fprintf(stderr,"File %s doesn't seem to contain patch information",path);
      exit(-1);
      }

   lseek(fd,pos,SEEK_SET);
   write(fd,mh_patchblock,sizeof(mh_patchblock));
   close(fd);
   }

}

